;; This version for simulator - uses initnetd.lisp  ** DOUBLE PRECISION **
;; Load flow calculation in Cartesian coordinates - corrected equations
;; Each processor plays three roles:
;;    1) Calculates c, d and sends to partner at other end of wire
;;    2) Recieves c, d from partner and participates in sums for C, D
;;    3) Represents its node, calculates new u, v (redundantly)

;; Solves quadratic system at each iteration, then uses overrelaxation to
;; form next approximation

(*proclaim '(type single-float WW))

(defvar WW 1.0)            ; The relaxation parameter, W > 1 overrelaxes

;; Additional working variables for solution calculation:

(*proclaim
 '(type double-float-pvar alfa beta gmma dlta Aq Bq Cq))

(*defvar alfa (!! 0.0))    ; alfa = f*C - a*D
(*defvar beta (!! 0.0))    ; beta = f*D + a*C
(*defvar gmma (!! 0.0))    ; w = gmma*v + dlta in solution
(*defvar dlta (!! 0.0))    ; 
(*defvar Aq   (!! 0.0))    ; At every node, quadratic eqn is
(*defvar Bq   (!! 0.0))    ;    Aq*v^2 + Bq^v + Cq = 0
(*defvar Cq   (!! 0.0))    ; 

;; Perform one iteration toward solution.  Returns maximum absolute
;; change in v, w components (after relaxtion)
;; This version assumes that D is never 0 in a generator.

(*proclaim '(type double-float-pvar oldv oldw))

(*defvar oldv (!! 0.0))   ; For convergence test
(*defvar oldw (!! 0.0))

(*proclaim '(type single-float maxch))

(defvar maxch 0.0)        ; maximum change in last iteration

(defun iterx ()
  (*set oldv v) (*set oldw w)
  (*when act
    (*set cm (-!! (*!! gm v) (*!! bm w)))
    (*set dm (+!! (*!! gm w) (*!! bm v)))
    (*pset :no-collisions cm C to)  (*pset :no-collisions dm D to)
    (*set C (reduce-and-spread!! C '+!! 1))
    (*set D (reduce-and-spread!! D '+!! 1))
    (*set alfa (-!! (*!! f C) (*!! a D)))
    (*set beta (+!! (*!! f D) (*!! a C)))
    (*if gen
      (progn
        (*set wk (-!! (*!! a alfa) (*!! f beta)))
        (*set gmma (/!! (+!! (*!! f alfa) (*!! a beta)) wk))
        (*set dlta (/!! (-!! (*!! f u2) (*!! (+!! (*!! f f) (*!! a a)) p))
                        wk))) ; end gen case
      (progn
        (*set gmma (/!! beta alfa))
        (*set dlta (/!! (-!! (*!! f q) (*!! a p)) alfa))
      )) ; end *if
    (*set Aq (+!! (!! 1.0) (*!! gmma gmma)))
    (*set Bq (*!! (!! 2.0) gmma dlta))
    (*set Cq (*!! dlta dlta))
    (*if gen
      (*set Cq (-!! Cq u2))
      (progn (*set Bq (+!! Bq alfa (*!! beta gmma)))
             (*set Cq (-!! (+!! Cq (*!! beta dlta)) (*!! f p) (*!! a q)))))
  ;; Solve quadratic equation
    (*set wk (-!! (*!! Bq Bq) (*!! (!! 4.0) Aq Cq)))
    (*set wk (sqrt!! wk))
    (*when clc  ;; Update all but slack node
      (*set v (/!! (-!! wk Bq) (*!! (!! 2.0) Aq)))
      (*set w (+!! (*!! gmma v) dlta))
      (*set wk (-!! w oldw))
      (*set w (+!! oldw (*!! (!! WW) wk)))  ;; Relaxed solution
      (*set wk (abs!! wk))
      (setq maxch (*max wk))
      (*set wk (-!! v oldv))
      (*set v (+!! oldv (*!! (!! WW) wk)))  ;; Relaxed solution
      (*set wk (abs!! wk))
      (setq maxch (max maxch (*max wk))))
  ) ; end *when act
  (* WW maxch)
)
